/* Copyright 2017 Alexander Debus
 *
 * This file is part of PIConGPU.
 *
 * PIConGPU is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PIConGPU is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PIConGPU.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "pmacc_types.hpp"
#include "simulation_defines.hpp"
#include "fields/laserProfiles/laserCylindricalGaussianHelper.hpp"

namespace picongpu
{
namespace laserCylindricalGaussian
{

    /**
     *
     * @param currentStep
     * @param subGrid
     * @param phase
     * @return
     */
    HINLINE float3_X laserLongitudinal( uint32_t const & currentStep, float_X & phase )
    {
        return float3_X::create( 0.0 );
    }

    /**
     *
     * @param elong
     * @param phase
     * @param posX
     * @param posZ
     * @param timestep
     * @return
     */
    HDINLINE float3_X laserTransversal(
        float3_X elong,
        float_X phase,
        float_X const posX,
        float_X const posZ,
        uint32_t const timestep
    )
    {
        PMacc::math::Vector< floatD_64, numComponents > const pos =
            getPositions(
                posX,
                posZ,
                false
            );

        float_X const focusPos = FOCUS_POS - laser::initPlaneY * CELL_HEIGHT;

        /* This is a work-around for a compiler bug of nvcc. Please check for new nvcc-versions if this is still required. */
        float_X const tmp_timeOffset = TIME_OFFSET;
        float_64 const tmp_Angle = INCIDENCE_ANGLE;

        float_64 const time = timestep * DELTA_T
            - precisionCast< float_64 >( focusPos / SPEED_OF_LIGHT )
            - precisionCast< float_64 >( tmp_timeOffset );
        float3_X result = float3_X::create( 0.0 );
        
        float3_64 fieldEPos( pos[ 0 ].x( ), pos[ 0 ].y( ), pos[ 0 ].x( ) );
        if( simDim == 3 )
        {
            fieldEPos.z( ) = pos[ 0 ].z( );
        }
        float_64 wfLongDist = calculateTemporalWavefrontDistance( fieldEPos );
        float_64 wfTransDist = calculateTransverseWavefrontDistance(
            wfLongDist,
            fieldEPos
        );
        result[ 0 ] = precisionCast< float_X >( calculateEx(
            time,
            fieldEPos
        ) * transverseFieldEnvelope( wfTransDist ) * longitudinalFieldEnvelope(
            wfLongDist,
            time
        ) );

        fieldEPos = float3_64( pos[ 1 ].x( ), pos[ 1 ].y( ), pos[ 1 ].x( ) );
        if( simDim == 3 )
        {
            fieldEPos.z( ) = pos[ 1 ].z( );
        }
        wfLongDist = calculateTemporalWavefrontDistance( fieldEPos );
        wfTransDist = calculateTransverseWavefrontDistance(
            wfLongDist,
            fieldEPos
        );
        result[ 1 ] = precisionCast< float_X >(
            (
                math::cos( tmp_Angle ) * calculateEy(
                    time,
                    fieldEPos
                ) - math::sin( tmp_Angle ) * calculateEz(
                    time,
                    fieldEPos
                )
            ) * transverseFieldEnvelope( wfTransDist ) * longitudinalFieldEnvelope(
                wfLongDist,
                time
            )
        );
        
        fieldEPos = float3_64( pos[ 2 ].x( ), pos[ 2 ].y( ), pos[ 2 ].x( ) );
        if( simDim == 3 )
        {
            fieldEPos.z( ) = pos[ 2 ].z( );
        }
        wfLongDist = calculateTemporalWavefrontDistance( fieldEPos );
        wfTransDist = calculateTransverseWavefrontDistance(
            wfLongDist,
            fieldEPos
        );
        result[ 2 ] = precisionCast< float_X >(
            (
                math::cos( tmp_Angle ) * calculateEz(
                    time,
                    fieldEPos
                ) + math::sin( tmp_Angle ) * calculateEy(
                    time,
                    fieldEPos
                )
            ) * transverseFieldEnvelope( wfTransDist ) * longitudinalFieldEnvelope(
                wfLongDist,
                time
            )
        );

        /* Note: This implementation deals with linear polarization in x only. */
        return AMPLITUDE * result; 
    }

} // namespace laserCylindricalGaussian
} // namespace picongpu
