/* Copyright 2017 Alexander Debus
 *
 * This file is part of PIConGPU.
 *
 * PIConGPU is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PIConGPU is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PIConGPU.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "pmacc_types.hpp"
#include "simulation_defines.hpp"
#include "math/Complex.hpp"
#include "math/Vector.hpp"
#include "dimensions/DataSpace.hpp"

namespace picongpu
{
namespace laserCylindricalGaussian
{

    HDINLINE PMacc::math::Complex< float_64 >
    calculateRefEx(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        /* Move this section into a more general area. */
        using complex_64 = PMacc::math::Complex< float_64 >;
        constexpr float_64 psi0 = float_64( 1.0 ); // This norm factor is itself not useful and only here for documentation.
        complex_64 const cii = complex_64( 0, 1 );
        float_64 const alpha = float_64( PI / 2. - INCIDENCE_ANGLE );
        float_64 const k = 2. * PI / WAVE_LENGTH;
        float_64 const omega = SPEED_OF_LIGHT * k;
        float_64 const kz = k * math::sin( alpha );
        float_64 const krho = k * math::cos( alpha );
        float_64 const xR = k * W0_X * W0_X / float_64( 2.0 );
        complex_64 const Xp = ( pos.x() + cii * xR );
        complex_64 const rhop = math::sqrt( Xp * Xp + pos.y() * pos.y() );

        complex_64 const Ex = cii * psi0 * math::exp( cii * ( kz * pos.z() - omega * time ) ) /
            (
                float_64( 2.0 ) * k * rhop * rhop * rhop * math::bessel::i0( krho * xR )
            ) *
            (
                rhop * ( k * kz * rhop * rhop - krho * krho * pos.y() * Xp ) * math::bessel::j0( krho * rhop ) +
                float_64( 2.0 ) * krho * pos.y() * Xp * math::bessel::j1( krho * rhop )
            );
         return Ex;
    }

    HDINLINE PMacc::math::Complex< float_64 >
    calculateRefEy(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        /* Move this section into a more general area. */
        using complex_64 = PMacc::math::Complex< float_64 >;
        constexpr float_64 psi0 = float_64( 1.0 ); // This norm factor is itself not useful and only here for documentation.
        complex_64 const cii = complex_64( 0, 1 );
        float_64 const alpha = float_64( PI / 2. - INCIDENCE_ANGLE );
        float_64 const k = 2. * PI / WAVE_LENGTH;
        float_64 const omega = SPEED_OF_LIGHT * k;
        float_64 const kz = k * math::sin( alpha );
        float_64 const krho = k * math::cos( alpha );
        float_64 const xR = k * W0_X * W0_X / float_64( 2.0 );
        complex_64 const Xp = ( pos.x() + cii * xR );
        complex_64 const rhop = math::sqrt( Xp * Xp + pos.y() * pos.y() );

        complex_64 const Ey = cii * psi0 * math::exp( cii * ( kz * pos.z() - omega * time ) ) /
            (
                float_64( 2.0 ) * k * rhop * rhop * rhop * math::bessel::i0( krho * xR )
            ) *
            (
                rhop * ( k * rhop - krho * pos.y() ) * ( k * rhop + krho * pos.y() ) * math::bessel::j0( krho * rhop ) -
                krho * ( rhop * rhop - float_64( 2.0 ) * pos.y() * pos.y() ) * math::bessel::j1( krho * rhop )
            );
        return Ey;
    }

    HDINLINE PMacc::math::Complex< float_64 >
    calculateRefEz(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        /* Move this section into a more general area. */
        using complex_64 = PMacc::math::Complex< float_64 >;
        constexpr float_64 psi0 = float_64( 1.0 ); // This norm factor is itself not useful and only here for documentation.
        complex_64 const cii = complex_64( 0, 1 );
        float_64 const alpha = float_64( PI / 2. - INCIDENCE_ANGLE );
        float_64 const k = 2. * PI / WAVE_LENGTH;
        float_64 const omega = SPEED_OF_LIGHT * k;
        float_64 const kz = k * math::sin( alpha );
        float_64 const krho = k * math::cos( alpha );
        float_64 const xR = k * W0_X * W0_X / float_64( 2.0 );
        complex_64 const Xp = ( pos.x() + cii * xR );
        complex_64 const rhop = math::sqrt( Xp * Xp + pos.y() * pos.y() );

        complex_64 const Ez = psi0 * math::exp( cii * ( kz * pos.z() - omega * time ) ) /
            (
                float_64( 2.0 ) * rhop * math::bessel::i0( krho * xR )
            ) *
            (
                krho * math::bessel::j1( krho * rhop ) * ( Xp + pos.y() * math::sin( alpha ) )
            );
        return Ez;
    }

    HDINLINE PMacc::math::Complex< float_64 >
    calculateRefHx(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        /* Move this section into a more general area. */
        using complex_64 = PMacc::math::Complex< float_64 >;
        constexpr float_64 psi0 = float_64( 1.0 ); // This norm factor is itself not useful and only here for documentation.
        constexpr float_64 sqrtEps = float_64( 1.0 ); // Square root of (rel.) dielectric constant in reference. For vacuum this is unity and is shown here for reference purposes only.
        complex_64 const cii = complex_64( 0, 1 );
        float_64 const alpha = float_64( PI / 2. - INCIDENCE_ANGLE );
        float_64 const k = 2. * PI / WAVE_LENGTH;
        float_64 const omega = SPEED_OF_LIGHT * k;
        float_64 const kz = k * math::sin( alpha );
        float_64 const krho = k * math::cos( alpha );
        float_64 const xR = k * W0_X * W0_X / float_64( 2.0 );
        complex_64 const Xp = ( pos.x() + cii * xR );
        complex_64 const rhop = math::sqrt( Xp * Xp + pos.y() * pos.y() );

        complex_64 const Hx = float_64( -1.0 ) * cii * psi0 * math::exp( cii * ( kz * pos.z() - omega * time ) ) * sqrtEps /
            (
                float_64( 2.0 ) * k * rhop * rhop * rhop * math::bessel::i0( krho * xR )
            ) *
            (
                rhop * ( k * kz * rhop * rhop + krho * krho * pos.y() * Xp ) * math::bessel::j0( krho * rhop ) -
                float_64( 2.0 ) * krho * pos.y() * Xp * math::bessel::j1( krho * rhop )
            );
        return Hx;
    }

    HDINLINE PMacc::math::Complex< float_64 >
    calculateRefHy(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        constexpr float_64 sqrtEps = float_64( 1.0 ); // Square root of (rel.) dielectric constant in reference. For vacuum this is unity and is shown here for reference purposes only.
        return sqrtEps * calculateRefEy( time, pos );
    }

    HDINLINE PMacc::math::Complex< float_64 >
    calculateRefHz(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        /* Move this section into a more general area. */
        using complex_64 = PMacc::math::Complex< float_64 >;
        constexpr float_64 psi0 = float_64( 1.0 ); // This norm factor is itself not useful and only here for documentation.
        constexpr float_64 sqrtEps = float_64( 1.0 ); // Square root of (rel.) dielectric constant in reference. For vacuum this is unity and is shown here for reference purposes only.
        complex_64 const cii = complex_64( 0, 1 );
        float_64 const alpha = float_64( PI / 2. - INCIDENCE_ANGLE );
        float_64 const k = 2. * PI / WAVE_LENGTH;
        float_64 const omega = SPEED_OF_LIGHT * k;
        float_64 const kz = k * math::sin( alpha );
        float_64 const krho = k * math::cos( alpha );
        float_64 const xR = k * W0_X * W0_X / float_64( 2.0 );
        complex_64 const Xp = ( pos.x() + cii * xR );
        complex_64 const rhop = math::sqrt( Xp * Xp + pos.y() * pos.y() );

        complex_64 const Hz = float_64( -1.0 ) * psi0 * math::exp( cii * ( kz * pos.z() - omega * time ) ) * sqrtEps /
            (
                float_64( 2.0 ) * rhop * math::bessel::i0( krho * xR )
            ) *
            (
                krho * math::bessel::j1( krho * rhop ) * ( Xp - pos.y() * math::sin( alpha ) )
            );
        return Hz;
    }

    /* Here we use included laser profile functions in order to calculate a non-trivial field normalization constant.
     * Reason: The reference normalizes one component of the vector potential and not the magnitude of the maximum possible electric field.
     * TO DO: Nicer placement of this definition. Requires definition of calculazeRefEx/Ey/Ez( ) functions
     * and is required by the calculateEx/Ey/Ez( ) functions.
     */
    //PMacc::math::Complex< float_64 > const ex0 = calculateRefEx( float_64( 0.0 ) , float3_64( 0.0, 0.0, 0.0 ) );
    //PMacc::math::Complex< float_64 > const ey0 = calculateRefEy( float_64( 0.0 ) , float3_64( 0.0, 0.0, 0.0 ) );
    //PMacc::math::Complex< float_64 > const ez0 = calculateRefEz( float_64( 0.0 ) , float3_64( 0.0, 0.0, 0.0 ) );
    //float_X const NORMALIZE_AMPLITUDE = precisionCast< float_X >( float_64( 1.0 ) /
    //    math::abs( math::sqrt( ex0 * ex0 + ey0 * ey0 + ez0 * ez0 ) ) );

    /* Hardcoded normalization factor for 800nm, W0_X=1.2e-6 and INCIDENCE_ANGLE = 5.0 degree */
    constexpr float_64 NORMALIZE_AMPLITUDE = 1.0 / ( 5.553135203593542e+6 * UNIT_LENGTH ) * float_64( 5.11574 ) * float_64( 739.609 ); // math::sqrt( float_64( 4.0 ) * PI * EPS0 );

    /** Wrapper function to cycle permutate position and field component coordinates
      *
      * Maps (ref Coordinates) x --> z (PIConGPU coordinates) and accordingly 
      * y->x, z->y. This mapping applies to the arguments of the reference functions,
      * as well as to the resulting fields.
      * Explanation: Polarization in y (Ref) becomes polarization in x (PIConGPU)
      * and focal line orientation in z (Ref) becomes focal line orientation in y (PIConGPU).
      * @param time of field evaluation
      * @param position for field evaluation
      * @return Ex field amplitude
      */
    HDINLINE float_64
    calculateEx(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        return math::sqrt( float_64( 4.0 ) * PI * EPS0 ) * NORMALIZE_AMPLITUDE * calculateRefEy(
            time,
            float3_64(
                pos.z(),
                pos.x(),
                pos.y()
            )
        ).get_real();
    }

    /** Wrapper function to cycle permutate position and field component coordinates
      *
      * Maps (ref Coordinates) x --> z (PIConGPU coordinates) and accordingly 
      * y->x, z->y. This mapping applies to the arguments of the reference functions,
      * as well as to the resulting fields.
      * Explanation: Polarization in y (Ref) becomes polarization in x (PIConGPU)
      * and focal line orientation in z (Ref) becomes focal line orientation in y (PIConGPU).
      * @param time of field evaluation
      * @param position for field evaluation
      * @return Ey field amplitude
      */
    HDINLINE float_64
    calculateEy(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        return math::sqrt( float_64( 4.0 ) * PI * EPS0 ) * NORMALIZE_AMPLITUDE * calculateRefEz(
            time,
            float3_64(
                pos.z(),
                pos.x(),
                pos.y()
            )
        ).get_real();
    }

    /** Wrapper function to cycle permutate position and field component coordinates
      *
      * Maps (ref Coordinates) x --> z (PIConGPU coordinates) and accordingly 
      * y->x, z->y. This mapping applies to the arguments of the reference functions,
      * as well as to the resulting fields.
      * Explanation: Polarization in y (Ref) becomes polarization in x (PIConGPU)
      * and focal line orientation in z (Ref) becomes focal line orientation in y (PIConGPU).
      * @param time of field evaluation
      * @param position for field evaluation
      * @return Ez field amplitude
      */
    HDINLINE float_64
    calculateEz(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        return math::sqrt( float_64( 4.0 ) * PI * EPS0 ) * NORMALIZE_AMPLITUDE * calculateRefEx(
            time,
            float3_64(
                pos.z(),
                pos.x(),
                pos.y()
            )
        ).get_real();
    }

    /** Wrapper function to cycle permutate position and field component coordinates
      *
      * Maps (ref Coordinates) x --> z (PIConGPU coordinates) and accordingly 
      * y->x, z->y. This mapping applies to the arguments of the reference functions,
      * as well as to the resulting fields.
      * Explanation: Polarization in y (Ref) becomes polarization in x (PIConGPU)
      * and focal line orientation in z (Ref) becomes focal line orientation in y (PIConGPU).
      * @param time of field evaluation
      * @param position for field evaluation
      * @return Bx field amplitude
      */
    HDINLINE float_64
    calculateBx(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        return math::sqrt( float_64( 4.0 ) * PI / MUE0 ) * NORMALIZE_AMPLITUDE * calculateRefHy(
            time,
            float3_64(
                pos.z(),
                pos.x(),
                pos.y()
            )
        ).get_real();
    }

    /** Wrapper function to cycle permutate position and field component coordinates
      *
      * Maps (ref Coordinates) x --> z (PIConGPU coordinates) and accordingly 
      * y->x, z->y. This mapping applies to the arguments of the reference functions,
      * as well as to the resulting fields.
      * Explanation: Polarization in y (Ref) becomes polarization in x (PIConGPU)
      * and focal line orientation in z (Ref) becomes focal line orientation in y (PIConGPU).
      * @param time of field evaluation
      * @param position for field evaluation
      * @return By field amplitude
      */
    HDINLINE float_64
    calculateBy(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        return math::sqrt( float_64( 4.0 ) * PI / MUE0 ) * NORMALIZE_AMPLITUDE * calculateRefHz(
            time,
            float3_64(
                pos.z(),
                pos.x(),
                pos.y()
            )
        ).get_real();
    }

    /** Wrapper function to cycle permutate position and field component coordinates
      *
      * Maps (ref Coordinates) x --> z (PIConGPU coordinates) and accordingly 
      * y->x, z->y. This mapping applies to the arguments of the reference functions,
      * as well as to the resulting fields.
      * Explanation: Polarization in y (Ref) becomes polarization in x (PIConGPU)
      * and focal line orientation in z (Ref) becomes focal line orientation in y (PIConGPU).
      * @param time of field evaluation
      * @param position for field evaluation
      * @return Bz field amplitude
      */
    HDINLINE float_64
    calculateBz(
        float_64 const & time,
        float3_64 const & pos
    )
    {
        return math::sqrt( float_64( 4.0 ) * PI / MUE0 ) * NORMALIZE_AMPLITUDE * calculateRefHx(
            time,
            float3_64(
                pos.z(),
                pos.x(),
                pos.y()
            )
        ).get_real();
    }

    HDINLINE float_64
    calculateTemporalWavefrontDistance( float3_64 const & pos )
    {
        /* Drop a perpendicular from X_0 to X_{center} (i.e. the surface of a cone
         * with fixed opening angle 5°) and find the intersection point.
         * X_{0,phi} + s * n_{cone surface normal,phi} == X_{center}
         * In reference coordinate system:
         * {x0, y0, z0} +  s {Cos[theta] Cos[phi], Cos[theta] Sin[phi], -Sin[theta]}
         *      ==  r {Sin[theta] Cos[phi], Sin[theta] Sin[phi], Cos[theta]}
         * Performed cyclical coordinate permutation on the result according to x->z, y->x, z->x
         * in order to make result correct for PIConGPU coordinate system.
         */
        float_64 const theta = float_64( INCIDENCE_ANGLE ); // This is THETA = INCIDENCE_ANGLE  = PI / 2 - ALPHA
        float_64 phi = PI;
        if( ( pos.z( ) != float_64( 0.0 ) ) || ( pos.x( ) != float_64( 0.0 ) ) )
            phi += math::atan2( pos.x( ), pos.z( ) );
        
        if( math::abs( pos.z( ) ) > math::abs( pos.x( ) ) )
        {
            return ( pos.y( ) + pos.z( ) / math::cos( phi ) * math::tan( theta ) ) /
                ( math::cos( theta ) + math::sin( theta ) * math::tan( theta ) );
        }
        else
        {
            return ( pos.y( ) + pos.x( ) / math::sin( phi ) * math::tan( theta ) ) /
                ( math::cos( theta ) + math::sin( theta ) * math::tan( theta ) );
        }
    }

    /**
     *
     * @param
     * @param
     * @return
     */
    HDINLINE float_64
    calculateTransverseWavefrontDistance(
        float_64 const & r,
        float3_64 const & pos
    )
    {
        /* Drop a perpendicular from X_0 to X_{center} (i.e. the surface of a cone
         * with fixed opening angle 5°) and find the intersection point.
         * X_{0,phi} + s * n_{cone surface normal,phi} == X_{center}
         * In reference coordinate system:
         * {x0, y0, z0} +  s {Cos[theta] Cos[phi], Cos[theta] Sin[phi], -Sin[theta]}
         *      ==  r {Sin[theta] Cos[phi], Sin[theta] Sin[phi], Cos[theta]}
         * Performed cyclical coordinate permutation on the result according to x->z, y->x, z->x
         * in order to make result correct for PIConGPU coordinate system.
         */
        float_64 const theta = float_64( INCIDENCE_ANGLE ); // This is THETA = INCIDENCE_ANGLE = PI / 2 - ALPHA
        float_64 phi = PI;
        if( ( pos.z( ) != float_64( 0.0 ) ) || ( pos.x( ) != float_64( 0.0 ) ) )
            phi += math::atan2( pos.x( ), pos.z( ) );
        float3_64 const intersection = r * float3_64(
            math::sin( theta ) * math::sin( phi ),
            math::cos( theta ),
            math::sin( theta ) * math::cos( phi )
        );
        return math::sqrt(
            ( intersection.x( ) - pos.x( ) ) * ( intersection.x( ) - pos.x( ) ) +
            ( intersection.y( ) - pos.y( ) ) * ( intersection.y( ) - pos.y( ) ) +
            ( intersection.z( ) - pos.z( ) ) * ( intersection.z( ) - pos.z( ) )
        );
    }

    HDINLINE float_X
    transverseFieldEnvelope( float_X const & distance )
    {
        if ( distance < PLATEAU_WIDTH )
        {
            return float_X( 1.0 );
        }
        else
        {
            return math::exp( -( ( distance - PLATEAU_WIDTH ) * ( distance - PLATEAU_WIDTH ) ) /
                ( float_X( 2.0 ) * SIGMA_TRANS * SIGMA_TRANS ) );
        }
    }
    
    HDINLINE float_X
    longitudinalFieldEnvelope(
        float_64 const & distance,
        float_64 const & time
    )
    {
        float_64 const diff = time - distance / SPEED_OF_LIGHT;
        return math::exp( -( diff * diff ) / ( PULSE_LENGTH * PULSE_LENGTH ) );
    }

    template <
        typename T_Type,
        typename T_AngleType
    >
    struct RotateVector;

    template <
        typename T_Type,
        typename T_AngleType
    >
    struct RotateVector<
        PMacc::math::Vector<
            T_Type,
            3
        >,
        T_AngleType
    >
    {
        using result = PMacc::math::Vector<
            T_Type,
            3
        >;
        using AngleType = T_AngleType;

        HDINLINE result
        operator( )( result const & fieldPosVector,
                    AngleType const & phi ) const
        {
            /*  Since, the laser propagation direction encloses an angle of phi with the
             *  simulation y-axis (i.e. direction of sliding window), the positions vectors are
             *  rotated around the simulation x-axis before calling the field calculation functions.
             *  
             *  Note: The field functions are in non-rotated frame and only use the angle
             *  phi to determine the required amount of pulse front tilt.
             *  Mathematica-Syntax: RotationMatrix[phi,{1,0,0}].{x,y,z}
             */
            return result(
                fieldPosVector.x( ),
                math::cos( phi ) * fieldPosVector.y( ) -
                    math::sin( phi ) * fieldPosVector.z( ),
                math::cos( phi ) * fieldPosVector.z( ) +
                    math::sin( phi ) * fieldPosVector.y( )
            );
        }
    };

    template <
        typename T_Type,
        typename T_AngleType
    >
    struct RotateVector<
        PMacc::math::Vector<
            T_Type,
            2
        >,
        T_AngleType
    >
    {
        using result = PMacc::math::Vector<
            T_Type,
            2
        >;
        using AngleType = T_AngleType;

        HDINLINE result
        operator( )( const result & fieldPosVector,
                    const AngleType phi ) const
        {
            /*  Since, the laser propagation direction encloses an angle of phi with the
             *  simulation y-axis (i.e. direction of sliding window), the positions vectors are
             *  rotated around the simulation x-axis before calling the field calculation functions.
             *  
             *  Note: The field functions are in non-rotated frame and only use the angle
             *  phi to determine the required amount of pulse front tilt.
             *  Mathematica-Syntax: RotationMatrix[phi].{x,y}
             */

             return result(
                math::cos( phi ) * fieldPosVector.x( ) -
                    math::sin( phi ) * fieldPosVector.y( ) ,
                math::cos( phi ) * fieldPosVector.y( ) +
                    math::sin( phi ) * fieldPosVector.x( ) );
        }
    };

    template <
        typename T_Type,
        typename T_AngleType
    >
    HDINLINE typename RotateVector<
        T_Type,
        T_AngleType
    >::result
    rotateVector( T_Type const & fieldPosVector,
                  T_AngleType const & phi )
    {
        return RotateVector< T_Type, T_AngleType >( )( fieldPosVector, phi );
    }

    /* Number of field components used in the simulation. [Default: 3 for both 2D and 3D] */
    constexpr uint32_t numComponents = 3u;

    HDINLINE PMacc::math::Vector<floatD_64, numComponents>
    getPositions( float_X const & posX, float_X const & posZ, bool const & isBField );

    HDINLINE PMacc::math::Vector<floatD_64, numComponents>
    getPositions( DataSpace< simDim > const & cellIdx, DataSpace< simDim > const & halfSimSize, bool const & isBField );

} // namespace laserCylindricalGaussian
} // namespace picongpu
