/* Copyright 2017 Alexander Debus
 *
 * This file is part of PIConGPU.
 *
 * PIConGPU is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PIConGPU is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PIConGPU.
 * If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "pmacc_types.hpp"
#include "simulation_defines.hpp"
#include "math/Complex.hpp"
#include "math/Vector.hpp"
#include "dimensions/DataSpace.hpp"

namespace picongpu
{
namespace laserCylindricalGaussian
{

    HDINLINE PMacc::math::Vector<floatD_64, numComponents>
    getPositions( float_X const & posX, float_X const & posZ, bool const & isBField = false )
    {
        float_X const posY = float_X( 0.0 );
        float3_X const pos0( posX, posY, posZ );
        floatD_64 const pos = precisionCast< float_64 >( pos0.shrink< simDim >( ) );
        
        /* For staggered fields (e.g. Yee-grid), obtain the fractional cell index components and add
         * that to the total cell indices. The physical field coordinate origin is transversally
         * centered with respect to the global simulation volume.
         * PMacc::math::Vector< floatD_X , numComponents > fieldGridPositions =
         *                fieldSolver::numericalCellType::traits::FieldPosition< FieldE >( );
         */
        fieldSolver::numericalCellType::traits::FieldPosition< FieldE > FieldPosE;
        fieldSolver::numericalCellType::traits::FieldPosition< FieldB > FieldPosB;
        PMacc::math::Vector<
            floatD_X,
            numComponents
        > fieldGridPositions = FieldPosE( );
        if ( isBField ) fieldGridPositions = FieldPosB( );
        
        floatD_64 const cellDimensions( picongpu::cellSize.shrink< simDim >( ) );
        float_X const focusPos = FOCUS_POS - laser::initPlaneY * CELL_HEIGHT;

        PMacc::math::Vector<
            floatD_64,
            numComponents
        > fieldPositions;

        /* This is a work-around for a compiler bug of nvcc. Please check for new nvcc-versions if this is still required. */
        float_64 const tmpIncidenceAngle = -INCIDENCE_ANGLE;

        for( uint32_t i = 0u; i < numComponents; ++i ) /* cellIdx Ex, Ey and Ez */
        {
            fieldPositions[i] = precisionCast< float_64 >( fieldGridPositions[ i ] ) * cellDimensions;
            fieldPositions[i] += pos;
            fieldPositions[i].y( ) -= precisionCast< float_64 >( focusPos ); // Translate back by initial distance from focus.
            fieldPositions[i] = rotateVector( fieldPositions[ i ], tmpIncidenceAngle );
        }

        return fieldPositions;
    }

    HDINLINE PMacc::math::Vector<floatD_64, numComponents>
    getPositions( DataSpace< simDim > const & cellIdx, DataSpace< simDim > const & halfSimSize, bool const & isBField = false )
    {
        /* Note: Neither direct precisionCast on picongpu::cellSize
           or casting on floatD_ does work. */
        floatD_64 const cellDim( picongpu::cellSize.shrink< simDim >( ) );

        /* TWTS laser coordinate origin is centered transversally and defined longitudinally by
           the laser center in y (usually maximum of intensity). */
        floatD_X laserOrigin = precisionCast< float_X >( halfSimSize );
        laserOrigin.y( ) = float_X( 0.0 );
        floatD_64 pos = floatD_64::create( 0.0 );
        pos = precisionCast< float_64 >( precisionCast< float_X >( cellIdx ) - laserOrigin ) * cellDim;

        /* For staggered fields (e.g. Yee-grid), obtain the fractional cell index components and add
         * that to the total cell indices. The physical field coordinate origin is transversally
         * centered with respect to the global simulation volume.
         * PMacc::math::Vector< floatD_X , numComponents > fieldGridPositions =
         *                fieldSolver::numericalCellType::traits::FieldPosition< FieldE >( );
         */
        fieldSolver::numericalCellType::traits::FieldPosition< FieldE > FieldPosE;
        fieldSolver::numericalCellType::traits::FieldPosition< FieldB > FieldPosB;
        PMacc::math::Vector<
            floatD_X,
            numComponents
        > fieldGridPositions = FieldPosE( );
        if ( isBField ) fieldGridPositions = FieldPosB( );

        floatD_64 const cellDimensions( picongpu::cellSize.shrink< simDim >( ) );
        float_X const focusPos = FOCUS_POS - laser::initPlaneY * CELL_HEIGHT;

        PMacc::math::Vector<
            floatD_64,
            numComponents
        > fieldPositions;

        /* This is a work-around for a compiler bug of nvcc. Please check for new nvcc-versions if this is still required. */
        float_64 const tmpIncidenceAngle = -INCIDENCE_ANGLE;

        for( uint32_t i = 0u; i < numComponents; ++i ) /* cellIdx Ex, Ey and Ez */
        {
            fieldPositions[i] = precisionCast< float_64 >( fieldGridPositions[ i ] ) * cellDimensions;
            fieldPositions[i] += pos;
            fieldPositions[i].y( ) -= precisionCast< float_64 >( focusPos ); // Translate back by initial distance from focus.
            fieldPositions[i] = rotateVector( fieldPositions[ i ], tmpIncidenceAngle );
        }

        return fieldPositions;
    }

} // namespace laserCylindricalGaussian
} // namespace picongpu
